/*
 * Test program to fit a 2D image to a Moffat function
 *
 * This program fit to a MOFFAT function 
 * an input FITs file!
 *
 */

#include <stdio.h> //For Standard I/O
#include <stdlib.h> //For Functions like ATOI
#include <ctype.h> //For Type check
#include <string.h> //For String manipulation
#include <math.h> //For Math functions
#include <time.h>

/*
 To use PGPLOT
 */

#include<cpgplot.h>

/*
To use the Cfitsio libraries
 */

#include<fitsio.h>

#include "my_math.h"
#include "my_struct.h"

//#include "principal_momentum.h"



#define MAX_SIZE 8192


#define SWAP(a,b) {temp=(a);(a)=(b);(b)=temp;}



float* readimage(char l_filename[80], long l_naxes[2]);
void writeimage(float* data, char l_filename[80], long l_naxes[2]);
void printerror(int status);

int main(int nargs,char* args[]);
double Model_xy(double x, double y, struct MODPAR *modpar, int n_mod);
double funk(double p[], struct IMG img, struct MODPAR *modpar, int n_mod, struct PSF psf);
int my_convolve(double *image_input, int nx_image_input, int ny_image_input, double *psf, int nx_psf, int ny_psf);

int read_config(char config[80], struct MODPAR *modpar, int *n_modpar, int *add_psf_out, double *ssq_out, double *min_delta_chisq);
int search_min(struct IMG img, struct MODPAR *modpar, int n_mod, struct PSF psf, int add_psf, double ssq, double *chisq, double min_delta_chisq);

int read_n_gauss(char config[80], struct MODPAR *modpar, int *n_modpar, int *add_psf_out, double *ssq_out, double *min_delta_chisq);
int write_n_gauss(char config[80], char kind[1], struct MODPAR *modpar, int n_mod);
int search_min_n_gauss(struct IMG img, struct MODPAR *modpar, int n_mod, double *chisq);


//void n_gauss(double x, double a[], double *y, double dyda[], int na);

#define MAX_MODELS 2000
int main(int nargs,char* args[]) {

  double chisq;
  struct MODPAR modpar[MAX_MODELS];
  struct PSF psf;
  struct IMG img;
  struct IMG *spec;

  char filename[80],psffilename[80],model[20],config[80];
  FILE *fichero;
  float *data,*model_data,*p_data;
  double data_temp,*dmodel;
  int naxes[2],psf_naxes[2];
  int i,k;
  double *a,*da;
  int *ia;
  double *x,*y;
  
  int nx,ny,nx_psf,ny_psf;

  double norm_psf;

  double min_delta_chisq;

  /*
   * For fitting
   */

  double *index,*flux,var[7][7],de,z,dz,p,*sigma;;
  double ssq,ssq2,d_ssq;
  int j;
  int np;
  double alamda;

  /*
   * Parameters for AMOEBA
   */
  double **p_in,*y_in;
  double *p_test,*p_p_test;
  int ndim;
  double ftol;
  int nfunk;
  int n_mod;
  int niter;
  int add_psf;

  double tmp_d;
  float tmp_f;
  int tmp_i;
  int aceptable;

  if (nargs<4) {
    printf("No arguments\n USE: data_file psf_file Config_File\n");
    return(0);
  } 



  strcpy(filename,args[1]);
  printf("Filename:%s\n",filename);
  data=readimage(filename,naxes);
  nx=naxes[0];
  ny=naxes[1];  
  printf("Read ok: Image Size [%d,%d]\n",nx,ny);
  img.nx=nx;
  img.ny=ny;
  spec = (struct IMG *)malloc(ny*sizeof(struct IMG));  
  
  
  /*
   * We build the x,y arrays
   */
  x=dvector(1,nx*ny);
  y=dvector(1,nx*ny);
  k=1;
  for (i=0;i<ny;i++) {
      spec[i].nx=nx;
      spec[i].ny=1;
      for (j=0;j<nx;j++) {
	  x[k]=(double)j+2;
	  y[k]=(double)i;
	  k++;
      }
  }
  printf("Position Ok\n");


  strcpy(psffilename,args[2]);
  printf("Filename:%s\n",psffilename);
  p_data=readimage(psffilename,psf_naxes);
  nx_psf=psf_naxes[0];
  ny_psf=psf_naxes[1];
  printf("Read ok: PSF Image Size [%d,%d]\n",nx_psf,ny_psf);
  psf.flux = (double *)malloc(nx_psf*ny_psf*sizeof(double));  
  norm_psf=0.0;
  for (j=0;j<(nx_psf*ny_psf);j++) {
      psf.flux[j]=(double)p_data[j];
      norm_psf+=psf.flux[j];
  }

  for (j=0;j<(nx_psf*ny_psf);j++) {
      psf.flux[j]/=norm_psf;
  }
  psf.npts=nx_psf*ny_psf; // 0 for no PSF correction
  psf.nx=nx_psf;
  psf.ny=ny_psf;



  /*
   * Defining the input values...
   */
  model_data = (float *)malloc(nx*ny*sizeof(float));  
  dmodel = (double *)malloc(nx*ny*sizeof(double));  
  img.flux= (double *)malloc(nx*ny*sizeof(double));  
  img.x= (double *)malloc(nx*ny*sizeof(double));  
  img.y= (double *)malloc(nx*ny*sizeof(double));  
  img.sigma= (double *)malloc(nx*ny*sizeof(double));  
  for (i=0;i<ny;i++) {
      spec[i].flux=(double *)malloc(nx*sizeof(double));  
      spec[i].x=(double *)malloc(nx*sizeof(double));  
      spec[i].sigma=(double *)malloc(nx*sizeof(double));  
  }


  /*
   * We feed the data!
   */
  index=dvector(1,nx*ny);
  flux=dvector(1,nx*ny);
  sigma=dvector(1,nx*ny);
  for (j=0;j<(nx*ny);j++) {
    index[j+1]=j*1.0;    
    flux[j+1]=(double)data[j];    
    img.flux[j]=(double)data[j];    
    sigma[j+1]=0.000001+sqrt(sqrt(flux[j+1]*flux[j+1]));
    img.sigma[j]=sigma[j+1];    
  }

  for (i=0;i<ny;i++) {
      for (j=0;j<nx;j++) {
	  img.x[j+i*nx]=(double)j+1;    
	  img.y[j+i*ny]=(double)i;    

	  spec[i].flux[j+1]=(double)data[j+i*ny];
	  spec[i].x[j]=(double)j+1;
	  spec[i].sigma[j]=0.000001+sqrt(sqrt(data[j+i*ny]));
      }
  }
  
  
  /*
   * Reading the config file
   *
   */
  //
  strcpy(config,args[3]);

/*

  printf("Reading N_GAUSS");
  read_n_gauss(config,modpar,&n_mod,&add_psf,&ssq,&min_delta_chisq);
  printf("Starting fitting\n");  
  search_min_n_gauss(img,modpar,n_mod,&chisq);
  printf("Fitting ended\n");
*/


  read_config(config,modpar,&n_mod,&add_psf,&ssq,&min_delta_chisq);
  printf("Starting fitting\n");
//  for (i=0;i<ny;i++) {
//      search_min(spec[i],modpar,n_mod,psf,add_psf,ssq,&chisq,min_delta_chisq);
//  }

      search_min(img,modpar,n_mod,psf,add_psf,ssq,&chisq,min_delta_chisq);
/*
  for (i=0;i<n_mod;i++) {
      modpar[i].ia[1]=0.01;
      modpar[i].p_a[3]=0.05;
      modpar[i].ia[4]=0;
      min_delta_chisq=0.1;
      ssq=0.1;
  }  
  printf("\nFirst Guess\n");
  search_min(img,modpar,n_mod,psf,add_psf,ssq,&chisq,min_delta_chisq);
*/
  for (j=0;j<ny;j++) {
      for (i=0;i<nx;i++) {
	  dmodel[i+j*nx]=Model_xy((double)i+2,(double)j,modpar,n_mod);
      }
  }
  if (psf.npts!=0) {
    my_convolve(dmodel,nx,ny,psf.flux,psf.nx,psf.ny);
  }
  for (j=0;j<(nx*ny);j++) {     
    model_data[j]=(float)dmodel[j];
  }

  printf("\n Model created\n");
  printf("%d F(%f,%f)=%f\n",j,x[j],y[j],model_data[j]);
  
  writeimage(model_data,"test_fit.fits",naxes);
  

 printf("Ok\n");
  return(0);
}



/*
 This function reads an standard FITS file with
 only one HDU and returns an image structure.
 */
float* readimage(char l_filename[80], long l_naxes[2]) {
  /* void readimage(char l_filename[80], long l_naxes[2], float* data){ */  
  float* data;
  fitsfile *fptr;       /* pointer to the FITS file, defined in fitsio.h */
  int status,  nfound, anynull;
//  long naxes[2], fpixel, nbuffer, npixels, ii;
  long nbuffer, npixels, ii;
  
  float datamin, datamax, nullval;
  float *buffer;

  int contador=0;
  long fpixel[2],i,lpixel[2],initialpixel;
  long inc;

      int j;
  
  status = 0;

  if ( fits_open_file(&fptr, l_filename, READONLY, &status) )
    printerror( status );
  /* read the NAXIS1 and NAXIS2 keyword to get image size */
    if ( fits_read_keys_lng(fptr, "NAXIS", 1, 2, l_naxes, &nfound, &status) )
         printerror( status );

    /*
      We read only the central portion of the image
      with a size of MAX_SIZE X Max_size.
      Not to desbord the data...
     */

    fpixel[0]=1;
    fpixel[1]=0;

    if (l_naxes[0]>MAX_SIZE) {
      fpixel[0]=(int)(l_naxes[0]-MAX_SIZE)/2;
    }
    if (l_naxes[1]>MAX_SIZE) {     
      fpixel[1]=(int)(l_naxes[1]-MAX_SIZE)/2;
    }

    lpixel[0]=fpixel[0]+l_naxes[0];
    lpixel[1]=fpixel[1]+l_naxes[1];



    npixels  = l_naxes[0]*l_naxes[1]; /* number of pixels in the image */
    nullval  = 0;                /* don't check for null values in the image */


    initialpixel=fpixel[0]+fpixel[1]*l_naxes[0];
    ii=0;


    buffer = (float *)malloc(l_naxes[0]*sizeof(float));
    data = (float *)malloc(npixels*sizeof(float));
    while(npixels>0)  {
      nbuffer=npixels;
      if (npixels>l_naxes[0])
	nbuffer=l_naxes[0];
 
      if(fits_read_img(fptr,TFLOAT,initialpixel,nbuffer,&nullval,buffer,&anynull,&status))
	printerror(status);

      for (j=0;j<nbuffer;j++){

	data[ii+j]=buffer[j];
      }

      ii+=nbuffer;
      npixels-=nbuffer;
      initialpixel += l_naxes[0];
      }


    if ( fits_close_file(fptr, &status) )
      printerror( status );

    return data;

}

void writeimage(float* data, char filename[80], long naxes[2]) {
  fitsfile *fptr;
  int status, ii, jj;
  long fpixel,nelements,exposure;
  int bitpix=FLOAT_IMG;
  long naxis=2;
  float mindata;
  float *local_data[naxes[1]];


  local_data[0]= (float*) malloc(naxes[0]*naxes[1] * sizeof (float));
  
  for (jj=1;jj<naxes[1];jj++) {
    local_data[jj]=local_data[jj-1]+naxes[0];
  }
  
  for (jj=0;jj<naxes[1];jj++) {
    for (ii=0;ii<naxes[0];ii++) {
      local_data[jj][ii]=(float)data[ii+jj*naxes[0]];
    }
  }
  
  remove(filename);
  if(fits_create_file(&fptr,filename,&status))
    printerror(status);
  if (fits_create_img(fptr,bitpix,naxis,naxes,&status))
    printerror(status);

  fpixel=1;
  nelements=naxes[0]*naxes[1];
  
  if (fits_write_img(fptr,TFLOAT,fpixel,nelements,local_data[0],&status))
    printerror(status);

  if (fits_close_file(fptr,&status))
    printerror(status);

  free(local_data[0]);
}


void printerror(int status) {
  if (status) {
    fits_report_error(stderr,status);
    exit(status);
  }
  return;
}



