/*
 S.F.Sanchez (08.08.2003)

*/


#include <stdlib.h>
#include <stdio.h>
#include <ctype.h> //For Type check
#include <string.h> //For String manipulation
#include <math.h> //For Math functions

#include <IFU_io.h>

#include "my_math.h"
#include "my_struct.h"

/*
 To use PGPLOT
 */
#include <cpgplot.h>
/* PGPLOT definitions */
#ifndef EXIT_FAILURE
#define EXIT_FAILURE 1
#endif

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 1
#endif

#define PARAMETERS 3
#define MAX_MODELS 2000



int copy_modpar(struct MODPAR *newpar, struct MODPAR *modpar, int n_mod);

double poly_n(double x, double a[], int na) {
    int i;
    double t = x;      
    double Yi = 0;
    for (i=0;i<na;i++) {
	Yi+=a[i]*pow(t,i);
    }
    return Yi;
}


double gauss1d(double x, double I_m,double r_m, double fondo) {
      double t = x;
      double e1 = exp(-0.5*pow((t-r_m)/fondo,2));
      double Yi = I_m * e1;
      return Yi;
}

double model(double x, double y, struct MODPAR *modpar, int n_mod) {
    int k,j;
    double flux_tmp,flux;
    double *a;
    double x_c,y_c,pre_pa,pa,q;
    double Im,r_m,beta;
    double xp,yp,r;
    double c;
    int nnx;

    flux=0;
    for (k=0;k<n_mod;k++) {
	x_c=modpar[k].a[1];
	y_c=modpar[k].a[2];
	Im=modpar[k].a[3];
	r_m=modpar[k].a[4];
	beta=modpar[k].a[5];
	pre_pa=modpar[k].a[6];
	pa=(3.14159/180.0)*modpar[k].a[6];
	while (pa>2*3.14159) {
	    pa-=2*3.14159;
	}
	while (pa<0) {
	    pa+=2*3.14159;
	}
	q=modpar[k].a[7];
	c=modpar[k].a[8];
	xp=(x-x_c)*cos(pa)+(y-y_c)*sin(pa);
	yp=(y-y_c)*cos(pa)-(x-x_c)*sin(pa);
	r=pow((pow(xp,(c+2))+pow(yp/q,(c+2))),1/(c+2));

	if(strcmp(modpar[k].model, "gauss1d") == 0)
	    flux_tmp=gauss1d(x,modpar[k].a[2],modpar[k].a[1],modpar[k].a[3]);
	else if(strcmp(modpar[k].model, "moffat1d") == 0)
	    flux_tmp=modpar[k].a[2]*(pow((1+pow(((x-modpar[k].a[1])/(modpar[k].a[3])),2)),-modpar[k].a[4]));
	else if(strcmp(modpar[k].model, "moffat") == 0)
	    flux_tmp=Im/(pow((1+pow((r/r_m),2)),beta));
	else if(strcmp(modpar[k].model, "gauss") == 0)
	    flux_tmp=Im*exp(-pow((r/(2*r_m)),2));
	else if(strcmp(modpar[k].model, "disk") == 0)
	    flux_tmp=Im*exp(-(r/r_m));
	else if(strcmp(modpar[k].model, "vauc") == 0)
	    flux_tmp=Im*exp(-7.67*(pow(r/r_m,0.25)-1));
	else if(strcmp(modpar[k].model, "back") == 0)
	    flux_tmp=Im;
	else if (strcmp(modpar[k].model, "poly1d") == 0) {
	    for (j=1;j<9;j++) {
		flux_tmp+=modpar[k].a[j]*pow(x,j-1);
	    }
	} else {
	    flux_tmp=0.0;
	}
	flux+=flux_tmp;
//	printf("%f %f\n",x,flux_tmp);
    }
    return flux;
}


double dmodel(double x, double y, struct MODPAR *modpar, int n_mod, int i_d) {
    int i,k,j;
    double flux1,flux2,flux;
    double d;
    double *a;
    int ndim;
    struct MODPAR *newpar;
    
    newpar = (struct MODPAR *)calloc(n_mod,sizeof(struct MODPAR));  
    d=0.1;
    copy_modpar(newpar,modpar,n_mod);
    read_modpar(a,modpar,n_mod,&ndim);
    i=0;
    for (k=0;k<n_mod;k++) {
	for(j=1;j<=modpar[k].na;j++) {
	    if (modpar[k].ia[j]==1) {
		if (modpar[k].fixa[j]==-1) {
		    if (i==i_d) {
			newpar[k].a[j]+=d;
		    }
		    i++;
		}
	    }
	}
    }
    flux1=model(x,y,newpar,n_mod);
    flux2=model(x,y,modpar,n_mod);
    flux=(flux1-flux2)/d;
    free(a);

    for (i=0;i<n_mod;i++) {

	free_vector(newpar[i].a,1,9);
	free_ivector(newpar[i].ia,1,9);
	free_ivector(newpar[i].fixa,1,9);
	free_vector(newpar[i].p_a,1,9);
	free_vector(newpar[i].p_a2,1,9);
/*
	free(newpar[i].a);
	free(newpar[i].ia);
	free(newpar[i].p_a);
	free(newpar[i].p_a2);
*/
    }
    free(newpar);
    return flux;
}


int write_error_par(double **covar, struct MODPAR *modpar, int n_mod);

/*
 * This routine extract a 1D spectra, within the indexes x1 and x2
 * of a 2D spectra (img_in);
 */
int sec_spec(struct IMG img_in,struct IMG *img, int j, int x1, int x2, int flag) {
    int i,k,l,n;
    double *tmp;
    double st,t;
    if (x1<0) x1=0;
    if (x2>img_in.nx) x2=img_in.nx;

    n=(x2-x1);
    img->nx= n;
    img->ny= 1;
    img->x= (double *)calloc(n,sizeof(double));  
    img->y= (double *)calloc(n,sizeof(double));  
    img->flux= (double *)calloc(n,sizeof(double));  
    img->sigma= (double *)calloc(n,sizeof(double));  
    if (flag==0) {
	for (i=0;i<(img->nx);i++) {
	    img->flux[i] = img_in.flux[i+x1+j*img_in.nx];	    
	    img->sigma[i] = sqrt(fabs(img->flux[i]));
	    img->x[i]= img_in.x[i+x1+j*img_in.nx];	    
	    img->y[i] = (double)j;
	}   
    } else {
	if (flag==1) {
	    tmp= (double *)calloc(j,sizeof(double));  
	    for (i=0;i<(img->nx);i++) {
		img->x[i]= img_in.x[i+x1];	    
		img->y[i] = 0.0;
		for (k=0;k<j;k++) {
		    tmp[k]=img_in.flux[i+x1+k*img_in.nx];	    
		}
		mean_loop(tmp,j,2,2,&t,&st);
		img->flux[i] = t;
	    img->sigma[i] = sqrt(fabs(img->flux[i])+st*st);
	    }   	 
	    free(tmp);
	} else {
	    if (j>(img_in.ny-5)) {
		j=img_in.ny-5;
	    }
	    l=j+5;
	    tmp= (double *)calloc(5,sizeof(double));  
	    for (i=0;i<(img->nx);i++) {
		img->x[i]= img_in.x[i+x1+j*img_in.nx];	    
		img->y[i] = 0.0;
		for (k=j;k<l;k++) {
		    tmp[k]=img_in.flux[i+x1+k*img_in.nx];	    
		}
		mean_loop(tmp,1,1,3,&t,&st);
		img->flux[i] = t;
		img->sigma[i] = sqrt(fabs(img->flux[i])+st*st);
	    }   	 
	    free(tmp);
	    
	}
    }
    return(0);
}


/*
 * This routine copy a 2D spec to another 2D spec
 */
int copy_spec(struct IMG img_in,struct IMG *img) {
    int i,n;
    n=(img_in.nx)*(img_in.ny);
    img->nx= img_in.nx;
    img->ny= img_in.ny;
    img->x= (double *)calloc(n,sizeof(double));  
    img->y= (double *)calloc(n,sizeof(double));  
    img->flux= (double *)calloc(n,sizeof(double));  
    img->sigma= (double *)calloc(n,sizeof(double));  
    for (i=0;i<n;i++) {
	img->flux[i] = img_in.flux[i];	    
	img->sigma[i] = img_in.sigma[i];
	img->x[i]= img_in.x[i];
	img->y[i]= img_in.y[i];
    }   
    return(0);
}


/*
 * This routine includes a 1D spectra, within the indexes x1 and x2
 * in a 2D spectra
 */
int in_spec(struct IMG *img_in, struct IMG img, int j, int x1, int x2) {
    int i,n;
    int nx_old;
    nx_old=img_in->nx;
    n=(x2-x1);
    for (i=0;i<img.nx;i++) {
	img_in->flux[i+x1+j*nx_old] = img.flux[i];	    
    }   
    return(0);
}


int read_spectra(char* filename, struct IMG *img_out, double *start, double *delta, double *end) {
    IMAGE2D image;
    FILE *fichero;
    float temp_float;    
    int count_f,n,i,temp_int;
    double start_w,delta_w;
    double d_temp[10],dval;
    int i_temp[10];
    int npix[2];
    struct IMG img;
    int j;
    printf("Opening the file %s\n",filename);
    fflush(stdout);
    open_frame(&image,filename,"I");
    printf("Ok\n");

    start_w=0.0;
    delta_w=1.0;

    if(!RD_desc(&image,"NAXIS1",INT,1,i_temp)) {
	printf("No NAXIS1 definied, error\n");
	exit_session(0);
	return(0);
    }
    npix[0]=i_temp[0];
    
    if(!RD_desc(&image,"NAXIS2",INT,1,i_temp)) {
	printf("No NAXIS2 definied, error\n");
	exit_session(0);
	return(0);
    }
    npix[1]=i_temp[0];
    
    if(!RD_desc(&image,"CRVAL1",DOUBLE,1,d_temp)) {
	printf("No CRVAL1 definied, error\n");
    } else {
	start_w=d_temp[0];
    }

    if(!RD_desc(&image,"CDELT1",DOUBLE,1,d_temp)) {
	printf("No CRDELT0 definied, error\n");
    } else {
	delta_w=d_temp[0];
    }

    if(!RD_desc(&image,"CD1_1",DOUBLE,1,d_temp)) {
	printf("No CD1_1 definied, error\n");
    } else {
	delta_w=d_temp[0];
    }





    n=npix[0]*npix[1];
    img_out->nx= npix[0];
    img_out->ny= npix[1];
    img_out->x= (double *)calloc(n,sizeof(double));  
    img_out->y= (double *)calloc(n,sizeof(double));  
    img_out->flux= (double *)calloc(n,sizeof(double));  
    img_out->sigma= (double *)calloc(n,sizeof(double));  

    for (j=0;j<image.ny;j++) {
	for (i=0;i<image.nx;i++) {
	    dval = (double)RD_frame(&image,i,j);
	    img_out->flux[i+j*image.nx] = dval;	    
	    img_out->sigma[i+j*image.nx] = sqrt(fabs(dval));	    
	    img_out->x[i+j*image.nx] = start_w+delta_w*i;	    
	    img_out->y[i+j*image.nx] = j;
	}
    }
    
    close_frame(&image);
//    *img_out=img;
    *start=start_w;
    *delta=delta_w;
    *end=start_w+delta_w*npix[0];
    return(0);
}


int save_spectra(struct IMG img, char* filename, double start_w, double step_w) {
  IMAGE2D image;
  int npix[2];
  double start[2];
  double step[2];
  int i,j;
  npix[0]=img.nx;
  npix[1]=img.ny;
  start[0]=0.0;
  start[1]=0.0;
  step[0]=1.0;
  step[1]=1.0;
  create_frame(&image,filename,npix,start,step,FLOAT,"MAP",NULL);
  for (i=0;i<img.nx;i++) {
    for (j=0;j<img.ny;j++) {
      WR_frame(&image,i,j,img.flux[i+j*img.nx]);
      WR_qframe(&image,i,j,(long) 0);
    }
  }
  start[0]=start_w;
  step[0]=step_w;
  WR_desc(&image,"CRVAL1",DOUBLE,1,start);
  WR_desc(&image,"CD1_1",DOUBLE,1,step);
  WR_desc(&image,"CDELT1",DOUBLE,1,step);
  close_frame(&image);  
  return(0);
}




int read_data(char* filename, struct IMG *img_out) {
    FILE *fichero;
    float temp_float;    
    int count_f,n,i,temp_int;
    struct IMG img;
    printf("Reading data...\n");
    n=0;
    if ((fichero=fopen(filename,"r"))==NULL) {
	printf("Error opening file: %s\n",filename);
	return(-1);
    } else {
	while( fscanf(fichero, "%f", &temp_float) != EOF) {	       			   
	    count_f++;
	    if (count_f>2) {
		count_f=0;
		n++;
	    }
	}
    }
    rewind(fichero);
    printf("");
    img_out->nx= n;
    img_out->x= (double *)calloc(n,sizeof(double));  
    img_out->y= (double *)calloc(n,sizeof(double));  
    img_out->flux= (double *)calloc(n,sizeof(double));  
    img_out->sigma= (double *)calloc(n,sizeof(double));  

    for(i=0;i<n;i++) {
	fscanf(fichero, "%d", &temp_int);
	fscanf(fichero, "%f", &temp_float);
	img_out->x[i]=temp_float;
	img_out->y[i]=1.0;
	fscanf(fichero, "%f", &temp_float);
	img_out->flux[i]=temp_float;
	img_out->sigma[i]=sqrt(fabs(temp_float));
    }
    printf("File readed\n");

//    *img_out=img;
    return(0);
}


int read_modpar(double *a, struct MODPAR *modpar, int n_mod, int *ndim) {
    int k,i,j;
    i=0;
    for (k=0;k<n_mod;k++) {
	for(j=1;j<=modpar[k].na;j++) {
	    if (modpar[k].ia[j]==1) {
		if (modpar[k].fixa[j]==-1) {
		    i++;
		}
	    }
	}
    }
    a= (double *)calloc(i,sizeof(double));  
    i=0;
    for (k=0;k<n_mod;k++) {
	for(j=1;j<=modpar[k].na;j++) {
	    if (modpar[k].ia[j]==1) {
		if (modpar[k].fixa[j]==-1) {
		    a[i]=modpar[k].a[j];
		    i++;
		}
	    }
	}
    }
    *ndim=i;
    return(0);
}


int write_modpar(double *a, struct MODPAR *modpar, int n_mod, int *ndim) {
    int k,i,j;
    i=0;
    for (k=0;k<n_mod;k++) {
	for(j=1;j<=modpar[k].na;j++) {
	    if (modpar[k].ia[j]==1) {
		if (modpar[k].fixa[j]==-1) {
		    modpar[k].a[j]=a[i];
		    i++;
		} else {			    
		    modpar[k].a[j]=modpar[modpar[k].fixa[j]-1].a[j]+modpar[k].p_a[j]; 		    
		}
	    }
	}
    }
    *ndim=i;
    return(0);
}


int write_p_modpar(double *a, struct MODPAR *modpar, int n_mod, int *ndim) {
    int k,i,j;
    i=0;
    for (k=0;k<n_mod;k++) {
	for(j=1;j<=modpar[k].na;j++) {
	    if (modpar[k].ia[j]==1) {
		if (modpar[k].fixa[j]==-1) {
		    modpar[k].p_a[j]=a[i];
		    i++;
		} else {			    
		    modpar[k].p_a[j]=modpar[modpar[k].fixa[j]-1].p_a[j];
		}
	    }
	}
    }
    *ndim=i;
    return(0);
}



int copy_modpar(struct MODPAR *newpar, struct MODPAR *modpar, int n_mod) {
    int k,j;
    int np=9;
    for (k=0;k<n_mod;k++) {	
	strcpy(newpar[k].model,modpar[k].model);
	newpar[k].na=modpar[k].na;
	newpar[k].a=dvector(1,np);
	newpar[k].p_a=dvector(1,np);
	newpar[k].ia=ivector(1,np);
	newpar[k].fixa=ivector(1,np);
	for(j=1;j<=modpar[k].na;j++) {
	    newpar[k].a[j]=modpar[k].a[j];
	    newpar[k].p_a[j]=modpar[k].a[j];
	    newpar[k].ia[j]=modpar[k].ia[j];
	    newpar[k].fixa[j]=modpar[k].fixa[j];
	}
    }
    return(0);
}




int read_config(char config[80], struct MODPAR *modpar, int *n_modpar, int *add_psf_out, double *ssq_out, double *min_delta_chisq);



void function(double x, double y, double *flux, double dp[], struct MODPAR modpar[],int n_mod,int ma);

void my_mrqcof_xy(double x[], double x1[], double y[], double sig[], int ndata, 
		  double **alpha, double beta[], double *chisq,
	    void (*funcs)(double, double, double *, double [], struct MODPAR modpar[], int n_mod), 
		  struct MODPAR modpar[], int n_mod);

void my_mrqmin_xy( double x[], double x1[], double y[], double sig[], int ndata, 
		   double **covar, double **alpha, double *chisq,
		   void (*funcs)(double, double, double *, double [], struct MODPAR modpar[], int n_mod),
		   double *alamda, struct MODPAR modpar[], int n_mod);

int print_par(struct MODPAR *modpar, int n_mod);


struct SKYLINES {
    int n;
    double *w;
    double *s_w;
    double *f;
    double *back;
};



int read_skylines(char skyfile[80], struct SKYLINES *skylines) {  
    FILE *fichero;
    int n,i,count_f;
    float temp_float;
    
    printf("Reading Skylines File\n");
    n=0;
    count_f=0;
    if ((fichero=fopen(skyfile,"r"))==NULL) {
	printf("Error opening file: %s\n",skyfile);
	return(-1);
    } else {
	while( fscanf(fichero, "%f", &temp_float) != EOF) {	       			   
	    count_f++;	    
	    if (count_f>2) {
		count_f=0;
		n++;
	    }
	}
    }
    rewind(fichero);
    fflush(stdout);
    skylines->n=n;
    skylines->w= (double *)calloc(n,sizeof(double));  
    skylines->s_w= (double *)calloc(n,sizeof(double));  
    skylines->f= (double *)calloc(n,sizeof(double));  
    skylines->back= (double *)calloc(n,sizeof(double));  
    for(i=0;i<n;i++) {
	fscanf(fichero, "%f", &temp_float);
	skylines->w[i]=(double)temp_float;
	fscanf(fichero, "%f", &temp_float);
	skylines->s_w[i]=(double)temp_float;
	fscanf(fichero, "%f", &temp_float);
	skylines->f[i]=(double)temp_float;
	skylines->back[i]=0.0;
    }
    fclose(fichero);
    printf("file closed\n");
    return(0);
}

int build_config(struct SKYLINES skylines, int n_line, double min, double max, double flex, struct MODPAR *modpar, int *n_modpar, int *add_psf_out, double *ssq_out, double *min_delta_chisq, int flag) {
    int i,j,np;
    float s_min;
   
    s_min=sqrt(fabs(min));   
    //printf("%f %f %f %f \n",max,min,max-min,s_min);
    if (((fabs(max-min)>s_min)&&(flag==0))||(flag==1)) {
	*add_psf_out=0;
	*n_modpar=2;
	*ssq_out=0.1;
	*min_delta_chisq=0.000001;
	np=9;
	strcpy(modpar[0].model,"gauss1d");
	strcpy(modpar[1].model,"poly1d");
	strcpy(modpar[2].model,"gauss1d");	
	strcpy(modpar[3].model,"gauss1d");
	for(i=0;i<2;i++) {
	    modpar[i].a=dvector(1,np);
	    modpar[i].p_a=dvector(1,np);
	    modpar[i].p_a2=dvector(1,np);
	    modpar[i].ia=ivector(1,np);
	    modpar[i].fixa=ivector(1,np);
	    for (j=0;j<np;) {
		j++;
		modpar[i].a[j]=0.0; 
		modpar[i].p_a[j]=0.0; 
		modpar[i].p_a2[j]=0.0; 
		modpar[i].ia[j]=0; 
		modpar[i].fixa[j]=-1; 
	    }
	    modpar[i].na=np;    
	    
	    
	}
	
	modpar[0].a[1]=skylines.w[n_line];
	if (flag==0) {
	    modpar[0].p_a[1]=skylines.w[n_line]-5*skylines.s_w[n_line];
	    modpar[0].p_a2[1]=skylines.w[n_line]+5*skylines.s_w[n_line];
	    modpar[0].ia[1]=1; 
	} else {
	    modpar[0].p_a[1]=skylines.w[n_line]-0.5*skylines.s_w[n_line];
	    modpar[0].p_a2[1]=skylines.w[n_line]+0.5*skylines.s_w[n_line];
	    modpar[0].ia[1]=1; 
	}
	if (flag==0) {
	    modpar[0].a[3]=skylines.s_w[n_line];
	    modpar[0].p_a[3]=skylines.s_w[n_line]-0.1*skylines.s_w[n_line];
	    modpar[0].p_a2[3]=skylines.s_w[n_line]+3*skylines.s_w[n_line];
	} else {
	    modpar[0].a[3]=skylines.s_w[n_line];
	    modpar[0].p_a[3]=skylines.s_w[n_line]-0.5*skylines.s_w[n_line];
	    modpar[0].p_a2[3]=skylines.s_w[n_line]+0.5*skylines.s_w[n_line];
	}
	modpar[0].ia[3]=1; 	
	if (flag==0) {
	    modpar[0].a[2]=(max-min)*2.0;
	    modpar[0].p_a[2]=(max-min)*0.01;
	    modpar[0].p_a2[2]=(max-min)*2.0;
	    modpar[0].ia[2]=1; 
	} else {
	    modpar[0].a[2]=skylines.f[n_line]*0.8;
	    modpar[0].p_a[2]=skylines.f[n_line]-skylines.f[n_line]*0.6;
	    modpar[0].p_a2[2]=skylines.f[n_line]+skylines.f[n_line]*1.7;
	    modpar[0].ia[2]=1; 
	}
	
	if (flag==0) {
	    modpar[1].a[1]=min;
	    modpar[1].p_a[1]=0.0;
	    modpar[1].p_a2[1]=max+0.5*fabs(s_min);

	} else {
	    modpar[1].a[1]=min;
	    modpar[1].p_a[1]=min;
	    modpar[1].p_a2[1]=max+0.5*fabs(s_min);
	}

	modpar[1].ia[1]=1; 
	modpar[1].a[2]=0.0;
	modpar[1].p_a[2]=0.0;
	modpar[1].p_a2[2]=10;
	modpar[1].ia[2]=0; 


	return(0);
    } else {
	return(-1);
    }
}



int build_config_mof(struct SKYLINES skylines, int n_line, double min, double max, double flex, struct MODPAR *modpar, int *n_modpar, int *add_psf_out, double *ssq_out, double *min_delta_chisq, int flag) {
    int i,j,np;
    float s_min;
   
    s_min=sqrt(fabs(min));   
    //printf("%f %f %f %f \n",max,min,max-min,s_min);
    if (((fabs(max-min)>s_min)&&(flag==0))||(flag==1)) {
	*add_psf_out=0;
	*n_modpar=2;
	*ssq_out=0.1;
	*min_delta_chisq=0.000001;
	np=9;
	strcpy(modpar[0].model,"moffat1d");
	strcpy(modpar[1].model,"poly1d");
	for(i=0;i<2;i++) {
	    modpar[i].a=dvector(1,np);
	    modpar[i].p_a=dvector(1,np);
	    modpar[i].p_a2=dvector(1,np);
	    modpar[i].ia=ivector(1,np);
	    modpar[i].fixa=ivector(1,np);
	    for (j=0;j<np;) {
		j++;
		modpar[i].a[j]=0.0; 
		modpar[i].p_a[j]=0.0; 
		modpar[i].p_a2[j]=0.0; 
		modpar[i].ia[j]=0; 
		modpar[i].fixa[j]=-1; 
	    }
	    modpar[i].na=np;    
	    
	    
	}
	
	modpar[0].a[1]=skylines.w[n_line];
	if (flag==0) {
	    modpar[0].p_a[1]=skylines.w[n_line]-5*skylines.s_w[n_line];
	    modpar[0].p_a2[1]=skylines.w[n_line]+5*skylines.s_w[n_line];
	    modpar[0].ia[1]=1; 
	} else {
	    modpar[0].p_a[1]=skylines.w[n_line]-2*skylines.s_w[n_line];
	    modpar[0].p_a2[1]=skylines.w[n_line]+2*skylines.s_w[n_line];
	    modpar[0].ia[1]=1; 
	}
	if (flag==0) {
	    modpar[0].a[3]=skylines.s_w[n_line];
	    modpar[0].p_a[3]=skylines.s_w[n_line]-0.1*skylines.s_w[n_line];
	    modpar[0].p_a2[3]=skylines.s_w[n_line]+flex*skylines.s_w[n_line];
	} else {
	    modpar[0].a[3]=skylines.s_w[n_line];
	    modpar[0].p_a[3]=skylines.s_w[n_line]-0.5*skylines.s_w[n_line];
	    modpar[0].p_a2[3]=skylines.s_w[n_line]+0.5*skylines.s_w[n_line];
	}
	modpar[0].ia[3]=1; 	
	if (flag==0) {
	    modpar[0].a[2]=(max-min)*2.0;
	    modpar[0].p_a[2]=(max-min)*0.01;
	    modpar[0].p_a2[2]=(max-min)*2.0;
	    modpar[0].ia[2]=1; 
	} else {
	    modpar[0].a[2]=skylines.f[n_line];
	    modpar[0].p_a[2]=skylines.f[n_line]*0.9;
	    modpar[0].p_a2[2]=skylines.f[n_line]*1.1;
	    modpar[0].ia[2]=1; 
	}

	modpar[0].a[4]=2.5;
	modpar[0].p_a[4]=0.1;
	modpar[0].p_a2[4]=10.0;
	modpar[0].ia[4]=1; 
	
//	if (flag==0) {
	    modpar[1].a[1]=min;
	    modpar[1].p_a[1]=0.0;
	    modpar[1].p_a2[1]=max+0.5*fabs(s_min);
/*
	} else {
	    modpar[1].a[1]=skylines.back[n_line];
	    modpar[1].p_a[1]=skylines.back[n_line]*0;
	    modpar[1].p_a2[1]=skylines.back[n_line]*1.1;
	}
*/
	modpar[1].ia[1]=1; 
	modpar[1].a[2]=0.0;
	modpar[1].p_a[2]=0.0;
	modpar[1].p_a2[2]=10;
	modpar[1].ia[2]=0; 

	return(0);
    } else {
	return(-1);
    }
}



int main (int argc,char** argv){


    int flag=0;
    char datafile[80],configfile[80],outfile[80],skyfile[80];
    char **argval, **arglabel;

    struct SKYLINES skylines;

    struct MODPAR modpar[MAX_MODELS];
    int n_mod,i;
    FILE *fichero;
    char filename[80];
    int count_f=0;
    float temp_float;
    int j;
    
    struct IMG img;
    struct IMG spec;
    int add_psf;
    double ssq,min_delta_chisq,chisq;
    int n;
    double chi_sq,real_chi_sq;

    float *xp,*yp,*yp_model,*yp_res,*sigma_p;
    float r_min,r_max,y_min,y_max;

    double **covar,**alpha;
    double alamda;
    double *a;
    int ndim;
  
    int niter=0;

    double start_w,delta_w,end_w;

    double min,max;
    int n_line;
    int i_start,i_end;

    int n_row;

    char c_a;
    int vis=0;

    set_arglist("-if|inputfile none -of|outputfile out.fits -sky skylines.list -vis 0");

    init_session(argv,argc,&arglabel,&argval);
    set_control_level(WARNING);


    for (i=0; i<4; i++) {
	if ((strstr(arglabel[i],"-if"))||(strstr(arglabel[i],"-inputfile"))) {
	    strcpy(datafile,argval[i]);
	}
	if ((strstr(arglabel[i],"-of"))||(strstr(arglabel[i],"-outputfile"))) {
	    strcpy(outfile,argval[i]);
	}

	if (strstr(arglabel[i],"-vis")) {
	    vis=atoi(argval[i]);
	}

	if (strstr(arglabel[i],"-sky")) {
	    strcpy(skyfile,argval[i]);
	}
    }  

    read_spectra(datafile,&img,&start_w,&delta_w,&end_w);
    read_skylines(skyfile,&skylines);

    for (i=0;i<skylines.n;i++) {
	
	//skylines.s_w[i]=sqrt(skylines.s_w[i]*skylines.s_w[i]+delta_w*delta_w);
	skylines.s_w[i]=delta_w;
    }


/*
 * We move through all the sky lines
 */
    if (vis!=0) {
	cpgopen("/xwin");
	cpgask(0);
    }

/*
 * We first fit to the collapsed rows!
 */

    for (n_line=0;n_line<skylines.n;n_line++) {
	i_start=floor((skylines.w[n_line]-start_w)/delta_w)-30/delta_w;
	i_end=floor((skylines.w[n_line]-start_w)/delta_w)+30/delta_w;

	
	if ((skylines.w[n_line]>start_w)&&(skylines.w[n_line]<end_w)) { 
	    sec_spec(img,&spec,5,i_start,i_end,1);
	    min=10000;
	    max=-10000;
	    for (i=0;i<spec.nx;i++) {
		if (spec.flux[i]>max) max=spec.flux[i];
		if (spec.flux[i]<min) min=spec.flux[i];
	    }
	    if (vis!=0) {
		printf("MIN,MAX=%f,%f\n",min,max);
	    }





	    if (build_config(skylines,n_line,min,max,0.5,modpar,&n_mod,&add_psf,&ssq,&min_delta_chisq,0)!=-1) {
		if (vis!=0) {
		    print_par(modpar,n_mod);
		    printf("Builded\n");
		}

//		fflush(stdout);

		xp= (float *)calloc(spec.nx,sizeof(float));  
		yp= (float *)calloc(spec.nx,sizeof(float));  
		yp_model= (float *)calloc(spec.nx,sizeof(float));  
		yp_res= (float *)calloc(spec.nx,sizeof(float));  
		r_min=100000;
		r_max=-100000;
		y_min=100000;
		y_max=-100000;
		n=spec.nx;
		for (i=0;i<spec.nx;i++) {
		    xp[i]=(float)spec.x[i];
		    yp[i]=(float)spec.flux[i];      
		    if (xp[i]>r_max) r_max=xp[i];
		    if (yp[i]>y_max) y_max=yp[i]+1.0;
		    if (xp[i]<r_min) r_min=xp[i];
		    if (yp[i]<y_min) y_min=yp[i]-1.0;
		}    
		read_modpar(a,modpar,n_mod,&ndim);
		free(a);
		
		if (flag==0) {
		    covar=dmatrix(1,ndim+1,1,ndim+1);
		    alpha=dmatrix(1,ndim+1,1,ndim+1);
		    flag=1;
		}


		alamda=-1.0;

		
		if (vis!=0) {
		    cpgsci(1);
		    cpgslw(2);
		    cpgenv(r_min,r_max,y_min,y_max,0,0);
		    cpglab("Wavelength", "Intensity", "");
		    cpgsci(1);
		    cpgpt(n,xp,yp,2);  
		}

		chisq=10e6;
		chi_sq=11e6;
		while (((chisq>ssq)&&(fabs(chi_sq-chisq)>min_delta_chisq))||(niter<4)) {
		    alamda=-1.0;
		    chi_sq=chisq;
		    my_mrqmin_xy(spec.x,spec.y,spec.flux,spec.sigma,spec.nx-1,covar,alpha,&chisq,&function,&alamda,modpar,n_mod);
		    if (vis!=0) {
			printf("Iter=%d Chisq=%f (%f), %f (%f), a=%f\n",niter,chisq,ssq,chi_sq-chisq,min_delta_chisq,alamda);
		    }
		    niter++;

		}
		
		y_min=100000;
		y_max=-100000;
		for (i=0;i<spec.nx;i++) {
		    yp_model[i]=model(spec.x[i],spec.y[i],modpar,n_mod);    
		    yp_res[i]=model(spec.x[i],spec.y[i],modpar,n_mod-1);    
		 
		}
		modpar[1].a[1]=0.0;

		for (i=0;i<spec.nx;i++) {
		    yp_res[i]=spec.flux[i]-model(spec.x[i],spec.y[i],modpar,n_mod);    
//		    spec.flux[i]=yp_res[i];

		}
		if (vis!=0) {
		    cpgsls(2);
		    cpgsci(2);
		    cpgline(n,xp,yp_model);  
		    cpgsci(4);
		    cpgpt(n,xp,yp_res,3);  		
		    cpgsci(1);
		    cpgsls(1);
		}

		if (vis!=0) {		
		    printf("Fitted\n");
		    fflush(stdout);
		}
		real_chi_sq=chisq/(spec.nx-n_mod-1);
		if (vis!=0) {
		    printf("CHI_SQ=%f\n",real_chi_sq);
		    write_error_par(covar,modpar,n_mod);
		    print_par(modpar,n_mod);
		    printf("ALAMDA=%f, <RETURN>\n",alamda);
		    c_a=getc(stdin);
		}

		free(xp);
		free(yp);
		free(yp_res);
		free(yp_model);

		skylines.w[n_line]=modpar[0].a[1];
		skylines.s_w[n_line]=modpar[0].a[3];
		skylines.f[n_line]=modpar[0].a[2];
		skylines.back[n_line]=modpar[1].a[1];

		for (i=0;i<n_mod;i++) {
		    free_vector(modpar[i].a,1,9);
		    free_ivector(modpar[i].ia,1,9);
		    free_ivector(modpar[i].fixa,1,9);
		    free_vector(modpar[i].p_a,1,9);
		    free_vector(modpar[i].p_a2,1,9);

		}


//		free_dmatrix(covar,1,ndim+1,1,ndim+1);
//		free_dmatrix(alpha,1,ndim+1,1,ndim+1);
	
	    } else {
		if (vis!=0) {
		    printf("No line detected %d %f\n",n_line,skylines.w[n_line]);
		}
	    }
	    free(spec.x);
	    free(spec.y);
	    free(spec.flux);
	    free(spec.sigma);
	    
	}
    
    }


/*
 * Now we repeat for each row!
 */

    if (vis!=0) {
	printf("**** We start with all the lines ****\n, <RETURN>\n");
	c_a=getc(stdin);
    }


    for (n_row=0;n_row<img.ny;n_row++) {
	for (n_line=0;n_line<skylines.n;n_line++) {
	    i_start=floor((skylines.w[n_line]-start_w)/delta_w)-30/delta_w;
	    i_end=floor((skylines.w[n_line]-start_w)/delta_w)+30/delta_w;
	    
	    if ((skylines.w[n_line]>start_w)&&(skylines.w[n_line]<end_w)) { 
		sec_spec(img,&spec,n_row,i_start,i_end,0);
		min=10000;
		max=-10000;
		for (i=0;i<spec.nx;i++) {
		if (spec.flux[i]>max) max=spec.flux[i];
		if (spec.flux[i]<min) min=spec.flux[i];
	    }

	    if (build_config(skylines,n_line,min,max,0.8,modpar,&n_mod,&add_psf,&ssq,&min_delta_chisq,1)!=-1) {
		if (vis!=0) {
		    print_par(modpar,n_mod);
		    printf("Builded\n");
		}
		xp= (float *)calloc(spec.nx,sizeof(float));  
		yp= (float *)calloc(spec.nx,sizeof(float));  
		yp_model= (float *)calloc(spec.nx,sizeof(float));  
		yp_res= (float *)calloc(spec.nx,sizeof(float));  
		r_min=100000;
		r_max=-100000;
		y_min=100000;
		y_max=-100000;
		n=spec.nx;
		for (i=0;i<spec.nx;i++) {
		    xp[i]=(float)spec.x[i];
		    yp[i]=(float)spec.flux[i];      
		    if (xp[i]>r_max) r_max=xp[i];
		    if (yp[i]>y_max) y_max=yp[i]+1.0;
		    if (xp[i]<r_min) r_min=xp[i];
		    if (yp[i]<y_min) y_min=yp[i]-1.0;
		}    
		read_modpar(a,modpar,n_mod,&ndim);
		free(a);
//		covar=dmatrix(1,ndim+1,1,ndim+1);
//		alpha=dmatrix(1,ndim+1,1,ndim+1);
		
		alamda=-1.0;

		
		if (vis!=0) {
		    cpgsci(1);
		    cpgslw(2);
		    cpgenv(r_min,r_max,y_min,y_max,0,0);
		    cpglab("Wavelength", "Intensity", "");
		    cpgsci(1);
		    cpgpt(n,xp,yp,2);  
		}
		    chisq=10e6;
		    chi_sq=11e6;
		chisq=10e6;
		chi_sq=11e6;
		while (((chisq>ssq)&&(fabs(chi_sq-chisq)>min_delta_chisq))||(niter<4)) {
		    alamda=-1.0;
		    chi_sq=chisq;
		    my_mrqmin_xy(spec.x,spec.y,spec.flux,spec.sigma,spec.nx-1,covar,alpha,&chisq,&function,&alamda,modpar,n_mod);
		    if (vis!=0) {
			print_par(modpar,n_mod);
			printf("Iter=%d Chisq=%f (%f), %f (%f), a=%f\n",niter,chisq,ssq,chi_sq-chisq,min_delta_chisq,alamda);
		    }
		    niter++;

		}
		
		
		
		
		y_min=100000;
		y_max=-100000;
		for (i=0;i<spec.nx;i++) {
		    yp_model[i]=model(spec.x[i],spec.y[i],modpar,n_mod);    
		    yp_res[i]=model(spec.x[i],spec.y[i],modpar,n_mod-1);    
		 
		}
		modpar[1].a[1]=0.0;

		for (i=0;i<spec.nx;i++) {
		    yp_res[i]=spec.flux[i]-model(spec.x[i],spec.y[i],modpar,n_mod);    
		    spec.flux[i]=yp_res[i];

		}
		if (vis!=0) {
		    cpgsls(2);
		    cpgsci(2);
		    cpgline(n,xp,yp_model);  
		    cpgsci(4);
		    cpgpt(n,xp,yp_res,3);  		
		    cpgsci(1);
		    cpgsls(1);
		}

		if (vis!=0) {		
		    printf("Fitted\n");
		    fflush(stdout);
		}
		real_chi_sq=chisq/(spec.nx-n_mod-1);
		if (vis!=0) {
		    printf("CHI_SQ=%f\n",real_chi_sq);
		    write_error_par(covar,modpar,n_mod);
		    print_par(modpar,n_mod);
		    printf("ALAMDA=%f, <RETURN>\n",alamda);
		    c_a=getc(stdin);
		}



		
		/*
		 * We clean the emission line!
		 */ 

		if (alamda>-666) {
		    in_spec(&img,spec,n_row,i_start,i_end);		    
		}
		free(xp);
		free(yp);
		free(yp_res);
		free(yp_model);
	    for (i=0;i<n_mod;i++) {

/*
		    free(modpar[i].a);
		    free(modpar[i].ia);
		    free(modpar[i].fixa);
		    free(modpar[i].p_a);
		    free(modpar[i].p_a2);
*/

		free_vector(modpar[i].a,1,9);
		free_ivector(modpar[i].ia,1,9);
		free_ivector(modpar[i].fixa,1,9);
		free_vector(modpar[i].p_a,1,9);
		free_vector(modpar[i].p_a2,1,9);
	    }
/*
		free(covar);
		free(alpha);
*/


		
	    } else {
		if (vis!=0) {
		    printf("No line detected\n",n_line,skylines.w[n_line]);
		}
	    }
	    free(spec.x);
	    free(spec.y);
	    free(spec.flux);
	    free(spec.sigma);

	    
	}

	
    }
	printf("ROW %d DONE\n",n_row);
//    c_a=getc(stdin);
    }

    free_dmatrix(covar,1,ndim+1,1,ndim+1);
    free_dmatrix(alpha,1,ndim+1,1,ndim+1);
    if (vis!=0) {
	cpgend();
    }
    printf("Saving file %s\n",outfile);
    save_spectra(img,outfile,start_w,delta_w);
    printf("File saved\n");
    exit_session(0);
}

    
    void function(double x, double y, double *flux, double dp[], struct MODPAR modpar[],int n_mod, int ma) {
	double *a;
	int ndim,j;
	*flux=model(x,y,modpar,n_mod); 
	for (j=1;j<=ma;j++) {
	    dp[j]=dmodel(x,y,modpar,n_mod,j-1); 
    }
    }


int read_config(char config[80], struct MODPAR *modpar, int *n_modpar, int *add_psf_out, double *ssq_out, double *min_delta_chisq) {

  
  FILE *fichero;
  int i,k;
  int j;
  int np;
  double alamda;
  double **p_in,*y_in;
  double *p_test,*p_p_test;
  int ndim;
  double ftol;
  int nfunk;
  int niter;
  int add_psf;
  
  double ssq;
  double min_delta_chisq_in;
  double tmp_d;
  float tmp_f;

  int tmp_i;
  int aceptable;
  int n_mod;

  printf("Config File : %s\n",config);
  if ((fichero=fopen(config,"r"))==NULL) {
    printf("Error opening file: %s\n",config);
    return(-1);
  } else {    
    printf("File Opened...\n");
    fscanf(fichero, "%d", &add_psf);
    if ((add_psf!=0)&&(add_psf!=1)) {
      add_psf=0;
    }
    printf("Add PSF=%d\n",add_psf);
    fflush(stdout);
    fscanf(fichero, "%d", &n_mod);
    printf("Num. Models= %d\n",n_mod);
    fflush(stdout);
    fscanf(fichero, "%f", &tmp_f);
    ssq=(double)tmp_f;
    printf("Chisq Lim= %f\n",ssq);
    fscanf(fichero, "%f", &tmp_f);
    min_delta_chisq_in=(double)tmp_f;
    printf("Min Delta_Chisq= %f\n",min_delta_chisq_in);
    fflush(stdout);
    np=9;
    ndim=0;  
    for(i=0;i<n_mod;i++) {
      modpar[i].a=dvector(1,np);
      modpar[i].p_a=dvector(1,np);
      modpar[i].p_a2=dvector(1,np);
      modpar[i].ia=ivector(1,np);
      modpar[i].fixa=ivector(1,np);
      fscanf(fichero,"%s",modpar[i].model); 
      for (j=0;j<np;) {
	j++;
	fscanf(fichero,"%g",&tmp_f); 
	modpar[i].a[j]=(double)tmp_f; 
	fscanf(fichero,"%d",&tmp_i); 
	modpar[i].ia[j]=tmp_i; 
	fscanf(fichero,"%g",&tmp_f); 
	modpar[i].p_a[j]=(double)tmp_f; 
	fscanf(fichero,"%g",&tmp_f); 
	modpar[i].p_a2[j]=(double)tmp_f; 
	fscanf(fichero,"%d",&tmp_i); 
	modpar[i].fixa[j]=tmp_i; 
	if (modpar[i].ia[j]==1) {
	  ndim++;
	}
      }
     
      modpar[i].ia[np]=0;
      modpar[i].na=np;    
      
    }
  }
  fclose(fichero);
  printf("File Closed\n");
  fflush(stdout);    

  
  *n_modpar=n_mod;
  *add_psf_out=add_psf;
  *ssq_out=ssq;
  *min_delta_chisq=min_delta_chisq_in;

  return(0);
}

